# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parser for LevelDB Table (.ldb) files."""
from __future__ import annotations

from dataclasses import dataclass, field
import io
import os
from typing import BinaryIO, Iterable, Tuple

import snappy
import zstd

from dfindexeddb import utils


@dataclass
class LdbKeyValueRecord:
  """A leveldb table key-value record.

  Attributes:
    offset: the offset of the record.
    key: the key of the record.
    value: the value of the record.
    sequence_number: the sequence number of the record.
    type: the type of the record.
  """
  offset: int
  key: bytes
  value: bytes
  sequence_number: int
  type: int

  PACKED_SEQUENCE_AND_TYPE_LENGTH = 8
  SEQUENCE_LENGTH = 7
  TYPE_LENGTH = 1

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, block_offset: int, shared_key: bytes
  ) -> Tuple[LdbKeyValueRecord, bytes]:
    """Decodes a ldb key value record.

    Args:
      decoder: the leveldb decoder.
      block_offset: the block offset.
      shared_key: the shared key bytes.

    Returns:
      A tuple of the parsed LdbKeyValueRecord and the updated shared key bytes.
    """
    offset, shared_bytes = decoder.DecodeUint32Varint()
    _, unshared_bytes = decoder.DecodeUint32Varint()
    _, value_length = decoder.DecodeUint32Varint()
    _, key_delta = decoder.ReadBytes(unshared_bytes)
    _, value = decoder.ReadBytes(value_length)

    shared_key = shared_key[:shared_bytes] + key_delta
    key = shared_key[:-cls.PACKED_SEQUENCE_AND_TYPE_LENGTH]
    sequence_number = int.from_bytes(
        key[-cls.SEQUENCE_LENGTH:], byteorder='little', signed=False)
    key_type = shared_key[-cls.PACKED_SEQUENCE_AND_TYPE_LENGTH]

    return cls(offset + block_offset, key, value, sequence_number,
               key_type), shared_key


@dataclass
class LdbBlock:
  """A leveldb table block.

  Attributes:
    offset: the offset of the block.
    block_offset:
  """
  offset: int
  block_offset: int
  length: int
  data: bytes = field(repr=False)
  footer: bytes  # 5 bytes = 1 byte compressed flag + 4 bytes checksum.

  SNAPPY_COMPRESSED = 1
  ZSTD_COMPRESSED = 2
  RESTART_ENTRY_LENGTH = 4

  def IsSnappyCompressed(self) -> bool:
    """Returns true if the block is snappy compressed."""
    return self.footer[0] == self.SNAPPY_COMPRESSED

  def IsZstdCompressed(self) -> bool:
    """Returns true if the block is zstd compressed."""
    return self.footer[0] == self.ZSTD_COMPRESSED

  def GetBuffer(self) -> bytes:
    """Returns the block buffer, decompressing if required."""
    if self.IsSnappyCompressed():
      return snappy.decompress(self.data)
    if self.IsZstdCompressed():
      return zstd.decompress(self.data)
    return self.data

  def GetRecords(self) -> Iterable[LdbKeyValueRecord]:
    """Returns an iterator over the key value records in the block.

    Yields:
      LdbKeyValueRecords
    """
    # get underlying block content, decompressing if required
    buffer = self.GetBuffer()
    decoder = utils.LevelDBDecoder(io.BytesIO(buffer))

    # trailer of a block has the form:
    #    restarts: uint32[num_restarts]
    #    num_restarts: uint32
    decoder.stream.seek(-self.RESTART_ENTRY_LENGTH, os.SEEK_END)
    _, num_restarts = decoder.DecodeUint32()
    restarts_offset = (
        decoder.stream.tell()) - (num_restarts + 1) * self.RESTART_ENTRY_LENGTH

    decoder.stream.seek(restarts_offset)
    _, offset = decoder.DecodeUint32()
    decoder.stream.seek(offset)
    key = b''

    while decoder.stream.tell() < restarts_offset:
      key_value_record, key = LdbKeyValueRecord.FromDecoder(
          decoder, self.block_offset, key)
      yield key_value_record

    # TODO: parse trailer of the block for restart points (where the full
    # key is stored to allow for binary lookup).  It's not needed at this time
    # since we are sequentially iterating over the records in the block/file.


@dataclass
class BlockHandle:
  """A handle to a block in the ldb file.

  Attributes:
    offset: the offset of the block handle.
    block_offset: the offset of the block.
    length: the length of the block.
  """
  offset: int
  block_offset: int
  length: int

  BLOCK_TRAILER_SIZE = 5

  def Load(self, stream: BinaryIO) -> LdbBlock:
    """Loads the block data.

    Args:
      stream: the binary stream of the ldb file.

    Returns:
      a LdbBlock.

    Raises:
      ValueError: if it could not read all of the block or block footer.
    """
    stream.seek(self.block_offset, os.SEEK_SET)
    data = stream.read(self.length)
    if len(data) != self.length:
      raise ValueError('Could not read all of the block')

    footer = stream.read(self.BLOCK_TRAILER_SIZE)
    if len(footer) != self.BLOCK_TRAILER_SIZE:
      raise ValueError('Could not read all of the block footer')

    return LdbBlock(self.offset, self.block_offset, self.length, data, footer)

  @classmethod
  def FromStream(cls, stream: BinaryIO, base_offset: int = 0) -> BlockHandle:
    """Reads a block handle from a binary stream.

    Args:
      stream: the binary stream.
      base_offset: the base offset.

    Returns:
      A BlockHandle.
    """
    decoder = utils.LevelDBDecoder(stream)
    offset, block_offset = decoder.DecodeUint64Varint()
    _, length = decoder.DecodeUint64Varint()
    return cls(offset + base_offset, block_offset, length)


class LdbFileReader:
  """A leveldb table (.ldb or .sst) file reader.

  A LdbFileReader provides read-only sequential iteration of serialized
  structures in a leveldb ldb file.  These structures include:
  * blocks (LdbBlock)
  * records (LdbKeyValueRecord)
  """

  FOOTER_SIZE = 48
  MAGIC = b'\x57\xfb\x80\x8b\x24\x75\x47\xdb'

  def __init__(self, filename: str):
    """Initializes the LogFile.

    Args:
      filename: the .ldb filename.

    Raises:
      ValueError if the file has an invalid magic number at the end.
    """
    self.filename = filename
    with open(self.filename, 'rb') as fh:
      fh.seek(-len(self.MAGIC), os.SEEK_END)
      if fh.read(len(self.MAGIC)) != self.MAGIC:
        raise ValueError(f'Invalid magic number in {self.filename}')

      fh.seek(-self.FOOTER_SIZE, os.SEEK_END)
      # meta_handle, need to read first due to variable integers
      _ = BlockHandle.FromStream(fh)
      index_handle = BlockHandle.FromStream(fh)

      # self.meta_block = meta_handle.load(fh)  # TODO: support meta blocks
      self.index_block = index_handle.Load(fh)

  def GetBlocks(self) -> Iterable[LdbBlock]:
    """Returns an iterator of LdbBlocks.

    Yields:
      LdbBlock.
    """
    with open(self.filename, 'rb') as fh:
      for key_value_record in self.index_block.GetRecords():
        block_handle = BlockHandle.FromStream(
            io.BytesIO(key_value_record.value),
            base_offset=key_value_record.offset)
        yield block_handle.Load(fh)

  def GetKeyValueRecords(self) -> Iterable[LdbKeyValueRecord]:
    """Returns an iterator of LdbKeyValueRecords.

    Yields:
      LdbKeyValueRecords.
    """
    for block in self.GetBlocks():
      yield from block.GetRecords()

  def RangeIter(self) -> Iterable[Tuple[bytes, bytes]]:  #pylint: disable=C0103
    """Returns an iterator of key-value pairs.

    Yields:
      A tuple of key and value as bytes.
    """
    for record in self.GetKeyValueRecords():
      yield (record.key, record.value)



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parses Chromium IndexedDb structures."""
from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime
import io
from typing import Any, BinaryIO, Optional, Tuple, Type, TypeVar, Union

from dfindexeddb import errors
from dfindexeddb import utils
from dfindexeddb.indexeddb import blink
from dfindexeddb.indexeddb import definitions
from dfindexeddb.leveldb import ldb
from dfindexeddb.leveldb import log


T = TypeVar('T')


@dataclass
class KeyPrefix(utils.FromStreamMixin):
  """The IndexedDB key prefix.

  Attributes:
    offset: the offset of the key prefix.
    database_id: the database ID.
    object_store_id: the object store ID.
    index_id: the index ID.
  """
  offset: int = field(compare=False)
  database_id: int
  object_store_id: int
  index_id: int

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, base_offset: int = 0) -> KeyPrefix:
    """Decodes a KeyPrefix from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      the decoded KeyPrefix.

    Raises:
      ParserError: when there is an invalid database/object store/index ID
        length.
    """
    offset, raw_prefix = decoder.ReadBytes(1)

    database_id_length = (raw_prefix[0] & 0xE0 >> 5) + 1
    object_store_id_length = (raw_prefix[0] & 0x1C >> 2) + 1
    index_id_length = (raw_prefix[0] & 0x03) + 1

    if database_id_length < 1 or database_id_length > 8:
      raise errors.ParserError('Invalid database ID length')

    if object_store_id_length < 1 or object_store_id_length > 8:
      raise errors.ParserError('Invalid object store ID length')

    if index_id_length < 1 or index_id_length > 4:
      raise errors.ParserError('Invalid index ID length')

    _, database_id = decoder.DecodeInt(database_id_length)
    _, object_store_id = decoder.DecodeInt(object_store_id_length)
    _, index_id = decoder.DecodeInt(index_id_length)

    return cls(
        offset=base_offset + offset,
        database_id=database_id,
        object_store_id=object_store_id,
        index_id=index_id)

  def GetKeyPrefixType(self) -> definitions.KeyPrefixType:
    """Returns the KeyPrefixType.

    The KeyPrefixType is based on the database/object store/index ID values.

    Raises:
      ParserError: if the key prefix is unknown.
    """
    if not self.database_id:
      return definitions.KeyPrefixType.GLOBAL_METADATA
    if not self.object_store_id:
      return definitions.KeyPrefixType.DATABASE_METADATA
    if self.index_id == 1:
      return definitions.KeyPrefixType.OBJECT_STORE_DATA
    if self.index_id == 2:
      return definitions.KeyPrefixType.EXISTS_ENTRY
    if self.index_id == 3:
      return definitions.KeyPrefixType.BLOB_ENTRY
    if self.index_id >= 30:
      return definitions.KeyPrefixType.INDEX_DATA
    raise errors.ParserError(
        f'Unknown KeyPrefixType (index_id={self.index_id})')


@dataclass
class IDBKey(utils.FromStreamMixin):
  """An IDBKey.

  Attributes:
    offset: the offset of the IDBKey.
    type: the type of the IDBKey.
    value: the value of the IDBKey.
  """
  offset: int = field(compare=False)
  type: definitions.IDBKeyType
  value: Union[list, bytes, str, float, datetime, None]

  _MAXIMUM_DEPTH = 2000

  @classmethod
  def FromDecoder(
      cls,
      decoder: utils.LevelDBDecoder,
      base_offset: int = 0  #pylint: disable=unused-argument
  ) -> IDBKey:
    """Decodes an IDBKey from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder
      base_offset: the base offset.

    Returns:
      An IDBKey

    Raises:
      ParserError: on invalid IDBKeyType or invalid array length during
        parsing.
      RecursionError: if maximum depth encountered during parsing.
    """

    def RecursiveParse(depth: int) -> Tuple[
        int, definitions.IDBKeyType, Union[
            list, bytes, str, float, datetime, None]]:
      """Recursively parses IDBKeys.

      Args:
        depth: the current recursion depth.

      Returns:
        A tuple of the offset, the key type and the key value (where the value
          can be bytes, str, float, datetime or a list of these types).

      Raises:
        ParserError: on invalid IDBKeyType or invalid array length during
          parsing.
        RecursionError: if maximum depth encountered during parsing.
      """
      if depth == cls._MAXIMUM_DEPTH:
        raise RecursionError('Maximum recursion depth encountered during parse')
      offset, key_type_value = decoder.DecodeInt(1)
      key_type = definitions.IDBKeyType(key_type_value)

      if key_type == definitions.IDBKeyType.NULL:
        value = None
      elif key_type == definitions.IDBKeyType.ARRAY:
        _, length = decoder.DecodeVarint()
        if length < 0:
          raise errors.ParserError('Invalid length encountered')
        value = []
        while length:
          entry = RecursiveParse(depth + 1)
          value.append(entry[2])
          length -= 1
      elif key_type == definitions.IDBKeyType.BINARY:
        _, value = decoder.DecodeBlobWithLength()
      elif key_type == definitions.IDBKeyType.STRING:
        _, value = decoder.DecodeStringWithLength()
      elif key_type == definitions.IDBKeyType.DATE:
        _, raw_value = decoder.DecodeDouble()
        value = datetime.utcfromtimestamp(raw_value/1000.0)
      elif key_type == definitions.IDBKeyType.NUMBER:
        _, value = decoder.DecodeDouble()
      elif key_type == definitions.IDBKeyType.MIN_KEY:
        value = None
      else:
        raise errors.ParserError('Invalid IndexedDbKeyType')
      return offset, key_type, value

    offset, key_type, value = RecursiveParse(0)
    return cls(base_offset + offset, key_type, value)


@dataclass
class IDBKeyPath(utils.FromStreamMixin):
  """An IDBKeyPath.

  Arguments:
    offset: the offset of the IDBKeyPath.
    type: the IDBKeyPath type.
    value: the IDBKeyPath value.
  """
  offset: int = field(compare=False)
  type: definitions.IDBKeyPathType
  value: Union[str, list[str], None]

  @classmethod
  def FromDecoder(cls, decoder: utils.LevelDBDecoder,
                   base_offset: int = 0) -> IDBKeyPath:
    """Decodes an IDBKeyPath from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      An IDBKeyPath

    Raises:
      ParserError: on insufficient bytes or invalid array length during
        parsing.
    """
    buffer = decoder.stream.getvalue()  #pytype: disable=attribute-error
    if len(buffer) < 3:
      raise errors.ParserError('Insufficient bytes to parse.')

    if buffer[0:2] != b'\x00\x00':
      offset, value = decoder.DecodeString()
      return IDBKeyPath(offset, definitions.IDBKeyPathType.STRING, value)

    _, type_bytes = decoder.ReadBytes(3)
    key_path_type_byte = type_bytes[2]
    key_path_type = definitions.IDBKeyPathType(key_path_type_byte)

    if key_path_type == definitions.IDBKeyPathType.NULL:
      value = None
      offset = decoder.stream.tell()
    elif key_path_type == definitions.IDBKeyPathType.STRING:
      offset, value = decoder.DecodeStringWithLength()
    elif key_path_type == definitions.IDBKeyPathType.ARRAY:
      value = []
      offset, count = decoder.DecodeVarint()
      if count < 0:
        raise errors.ParserError(f'Invalid array length {count}')
      for _ in range(count):
        _, entry = decoder.DecodeStringWithLength()
        value.append(entry)
    return IDBKeyPath(base_offset + offset, key_path_type, value)


@dataclass
class BlobJournalEntry(utils.FromStreamMixin):
  """A blob journal entry.

  Attributes:
    offset (int): the offset.
    database_id (int): the database ID.
    blob_number (int): the blob number.
  """
  offset: int = field(compare=False)
  database_id: int
  blob_number: int

  @classmethod
  def FromDecoder(cls, decoder: utils.LevelDBDecoder,
                   base_offset: int = 0) -> BlobJournalEntry:
    """Decodes a BlobJournalEntry from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      A BlobJournalEntry.
    """
    offset, database_id = decoder.DecodeUint64Varint()
    _, blob_number = decoder.DecodeUint64Varint()
    return cls(offset=base_offset + offset, database_id=database_id,
               blob_number=blob_number)


@dataclass
class BlobJournal(utils.FromStreamMixin):
  """A BlobJournal.

  Attributes:
    offset (int): the offset.
    entries (list[BlobJournalEntry]): the list of blob journal entries.
  """
  offset: int = field(compare=False)
  entries: list[BlobJournalEntry]

  @classmethod
  def FromDecoder(cls, decoder: utils.LevelDBDecoder,
                   base_offset: int = 0) -> BlobJournal:
    """Decodes a BlobJournal from the current position of a LevelDBDecoder.

    Blob journals are zero-or-more instances of BlobJournalEntry.  There is no
    length prefix; just read until you run out of data.

    Ref: indexeddb_db_leveldb_coding.cc#DecodeBlobJournal

    Args:
      decoder: the LevelDBDecoder
      base_offset: the base offset.

    Returns:
      A BlobJournal
    """
    offset = decoder.stream.tell()
    entries = []

    # since there is no length prefix, consume the remaining buffer
    while True:
      try:
        journal_entry = BlobJournalEntry.FromDecoder(decoder, base_offset)
      except errors.DecoderError:
        break
      entries.append(journal_entry)

    return cls(offset=base_offset + offset, entries=entries)


@dataclass
class BaseIndexedDBKey:
  """A base class for IndexedDB coded leveldb keys and values.

  This class provides an interface to parse the key (using from_bytes/
  from_stream/from_decoder) and the value (using parse_value/decode_value).

  Attributes:
    offset: the offset of the key (after the key_prefix).
    key_prefix: the key prefix.
  """
  offset: int
  key_prefix: KeyPrefix

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> Any:
    """Decodes the value from the current position of the LevelDBDecoder.

    To be implemented by subclasses.

    Args:
      decoder: the stream decoder

    Raises:
      NotImplementedError.
    """
    raise NotImplementedError(f'{self.__class__.__name__}.decode_value')

  def ParseValue(self, value_data: bytes) -> Any:
    """Parses the value from raw bytes.

    Args:
      value_data: the raw value bytes.

    Returns:
      The parsed value.
    """
    if not value_data:
      return None
    decoder = utils.LevelDBDecoder(io.BytesIO(value_data))
    return self.DecodeValue(decoder)

  @classmethod
  def FromDecoder(
      cls: T,
      decoder: utils.LevelDBDecoder,
      key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> T:  #pylint: disable=unused-variable
    """Decodes the remaining key data from the current decoder position.

    Args:
      decoder: the stream decoder.
      key_prefix: the decoded key_prefix.
      base_offset: the base offset.

    Returns:
      The decoded key.

    Raises:
      NotImplementedError.
    """
    raise NotImplementedError(f'{cls.__class__.__name__}.decode_key')

  @classmethod
  def FromStream(
      cls: Type[T], stream: BinaryIO, base_offset: int = 0) -> T:
    """Parses the key from the current position of the binary stream.

    Args:
      stream: the binary stream.
      base_offset: the base offset of the stream.

    Returns:
      The decoded key.
    """
    decoder = utils.LevelDBDecoder(stream)
    key_prefix = KeyPrefix.FromDecoder(decoder, base_offset=base_offset)
    return cls.FromDecoder(
        decoder=decoder, key_prefix=key_prefix, base_offset=base_offset)

  @classmethod
  def FromBytes(
      cls: Type[T], raw_data: bytes, base_offset: int = 0) -> T:
    """Parses the key from the raw key bytes.

    Args:
      raw_data: the raw key data.
      base_offset: the base offset of the key data.

    Returns:
      The decoded key.
    """
    stream = io.BytesIO(raw_data)
    return cls.FromStream(stream=stream, base_offset=base_offset)


@dataclass
class SchemaVersionKey(BaseIndexedDBKey):
  """A schema version IndexedDb key."""

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> int:
    """Decodes the schema version value."""
    return decoder.DecodeVarint()[1]

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> SchemaVersionKey:
    """Decodes the schema version key."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.GlobalMetadataKeyType.SCHEMA_VERSION:
      raise errors.ParserError('Not a SchemaVersionKey')
    return cls(offset=base_offset + offset, key_prefix=key_prefix)


@dataclass
class MaxDatabaseIdKey(BaseIndexedDBKey):
  """A max database ID IndexedDB key."""

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> int:
    """Decodes the maximum database value."""
    return decoder.DecodeVarint()[1]

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> MaxDatabaseIdKey:
    """Decodes the maximum databse key."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.GlobalMetadataKeyType.MAX_DATABASE_ID:
      raise errors.ParserError('Not a MaxDatabaseIdKey')
    return cls(offset=base_offset + offset, key_prefix=key_prefix)


@dataclass
class DataVersionKey(BaseIndexedDBKey):
  """A data version IndexedDB key."""

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> int:
    """Decodes the data version value."""
    return decoder.DecodeUint64Varint()[1]

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> DataVersionKey:
    """Decodes the data version key."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.GlobalMetadataKeyType.DATA_VERSION:
      raise errors.ParserError('Not a DataVersionKey')
    return cls(offset=base_offset + offset, key_prefix=key_prefix)


@dataclass
class RecoveryBlobJournalKey(BaseIndexedDBKey):
  """A recovery blob journal IndexedDB key."""

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> BlobJournal:
    """Decodes the recovery blob journal value."""
    return BlobJournal.FromDecoder(decoder)

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> RecoveryBlobJournalKey:
    """Decodes the recovery blob journal key."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.GlobalMetadataKeyType.RECOVERY_BLOB_JOURNAL:
      raise errors.ParserError('Not a RecoveryBlobJournalKey')
    return cls(offset=base_offset + offset, key_prefix=key_prefix)


@dataclass
class ActiveBlobJournalKey(BaseIndexedDBKey):
  """An active blob journal IndexedDB key."""

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> BlobJournal:
    """Decodes the active blob journal value."""
    return BlobJournal.FromDecoder(decoder)

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> ActiveBlobJournalKey:
    """Decodes the active blob journal value."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.GlobalMetadataKeyType.ACTIVE_BLOB_JOURNAL:
      raise errors.ParserError('Not a ActiveBlobJournalKey')
    return cls(offset=base_offset + offset, key_prefix=key_prefix)


@dataclass
class EarliestSweepKey(BaseIndexedDBKey):
  """An earliest sweep IndexedDB key."""

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> int:
    """Decodes the earliest sweep value."""
    _, data = decoder.ReadBytes()
    return int.from_bytes(data, byteorder='little', signed=False)

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> EarliestSweepKey:
    """Decodes the earliest sweep value."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.GlobalMetadataKeyType.EARLIEST_SWEEP:
      raise errors.ParserError('Not a EarliestSweepKey')
    return cls(offset=base_offset + offset, key_prefix=key_prefix)


@dataclass
class EarlistCompactionTimeKey(BaseIndexedDBKey):
  """An earliest compaction time IndexedDB key."""

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> int:
    """Decodes the earliest compaction time value."""
    _, data = decoder.ReadBytes()
    return int.from_bytes(data, byteorder='little', signed=False)

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> EarlistCompactionTimeKey:
    """Decodes the earliest compaction time key."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.GlobalMetadataKeyType.EARLIEST_COMPACTION_TIME:
      raise errors.ParserError('Not a EarlistCompactionTimeKey')
    return cls(offset=base_offset + offset, key_prefix=key_prefix)


@dataclass
class ScopesPrefixKey(BaseIndexedDBKey):
  """A scopes prefix IndexedDB key."""

  def DecodeValue(self, decoder: utils.StreamDecoder) -> Optional[bytes]:
    """Decodes the scopes prefix value."""
    if decoder.NumRemainingBytes:
      return decoder.ReadBytes()[1]
    return None

  @classmethod
  def FromDecoder(
      cls, decoder: utils.StreamDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> ScopesPrefixKey:
    """Decodes the scopes prefix key."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.GlobalMetadataKeyType.SCOPES_PREFIX:
      raise errors.ParserError('Not a ScopesPrefixKey')
    return cls(offset=base_offset + offset, key_prefix=key_prefix)


@dataclass
class DatabaseFreeListKey(BaseIndexedDBKey):
  """A database free list IndexedDB key.

  Attributes:
    database_id: the database ID.
  """
  database_id: int

  def DecodeValue(self, decoder: utils.LevelDBDecoder):
    """Decodes the database free list value."""
    raise NotImplementedError('DatabaseFreeListKey.decode_value')

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> DatabaseFreeListKey:
    """Decodes the database free list key."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.GlobalMetadataKeyType.DATABASE_FREE_LIST:
      raise errors.ParserError('Not a DatabaseFreeListKey')
    _, database_id = decoder.DecodeVarint()
    return cls(
        offset=base_offset + offset, key_prefix=key_prefix,
        database_id=database_id)


@dataclass
class DatabaseNameKey(BaseIndexedDBKey):
  """A database name key.

  Attributes:
    origin: the origin of the database.
    database_name: the database name.
  """
  origin: str
  database_name: str

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> int:
    """Decodes the database name value.

    The value is the corresponding database ID.
    """
    return decoder.DecodeVarint()[1]

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> DatabaseNameKey:
    """Decodes the database name key."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.GlobalMetadataKeyType.DATABASE_NAME:
      raise errors.ParserError('Not a DatabaseNameKey')
    _, origin = decoder.DecodeStringWithLength()
    _, database_name = decoder.DecodeStringWithLength()
    return cls(
        offset=base_offset + offset, key_prefix=key_prefix, origin=origin,
        database_name=database_name)


@dataclass
class GlobalMetaDataKey(BaseIndexedDBKey):
  """A GlobalMetaDataKey parser."""

  METADATA_TYPE_TO_CLASS = {
      definitions.GlobalMetadataKeyType
          .SCHEMA_VERSION: SchemaVersionKey,
      definitions.GlobalMetadataKeyType
          .MAX_DATABASE_ID: MaxDatabaseIdKey,
      definitions.GlobalMetadataKeyType
          .DATA_VERSION: DataVersionKey,
      definitions.GlobalMetadataKeyType
          .RECOVERY_BLOB_JOURNAL: RecoveryBlobJournalKey,
      definitions.GlobalMetadataKeyType
          .ACTIVE_BLOB_JOURNAL: ActiveBlobJournalKey,
      definitions.GlobalMetadataKeyType
          .EARLIEST_SWEEP: EarliestSweepKey,
      definitions.GlobalMetadataKeyType
          .EARLIEST_COMPACTION_TIME: EarlistCompactionTimeKey,
      definitions.GlobalMetadataKeyType
          .SCOPES_PREFIX: ScopesPrefixKey,
      definitions.GlobalMetadataKeyType
          .DATABASE_FREE_LIST: DatabaseFreeListKey,
      definitions.GlobalMetadataKeyType
          .DATABASE_NAME: DatabaseNameKey}

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> Any:
    """Decodes the value from the current position of the LevelDBDecoder.

    Args:
      decoder: the stream decoder
    """

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> Union[Type[ActiveBlobJournalKey],
             Type[DataVersionKey],
             Type[DatabaseFreeListKey],
             Type[DatabaseNameKey],
             Type[EarliestSweepKey],
             Type[EarlistCompactionTimeKey],
             Type[MaxDatabaseIdKey],
             Type[RecoveryBlobJournalKey],
             Type[SchemaVersionKey],
             Type[ScopesPrefixKey]]:
    """Decodes the global metadata key.

    Raises:
      ParserError: if the key contains an unknown metadata key type.
    """
    _, metadata_value = decoder.PeekBytes(1)
    metadata_type = definitions.GlobalMetadataKeyType(metadata_value[0])

    key_class = cls.METADATA_TYPE_TO_CLASS.get(metadata_type)
    if not key_class:
      raise errors.ParserError('Unknown metadata key type')
    return key_class.FromDecoder(
        decoder, key_prefix, base_offset)  #pytype: disable=bad-return-type


@dataclass
class ObjectStoreFreeListKey(BaseIndexedDBKey):
  """An IndexedDB object store free list key.

  Attributes:
    object_store_id: the ID of the object store containing the free list.
  """
  object_store_id: int

  def DecodeValue(self, decoder: utils.LevelDBDecoder):
    """Decodes the object store free list value."""
    raise NotImplementedError('ObjectStoreFreeListKey.decode_value')

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> ObjectStoreFreeListKey:
    """Decodes the object store free list key."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.DatabaseMetaDataKeyType.OBJECT_STORE_FREE_LIST:
      raise errors.ParserError('Not am ObjectStoreFreeListKey')
    _, object_store_id = decoder.DecodeVarint()
    return cls(
        offset=base_offset + offset, key_prefix=key_prefix,
        object_store_id=object_store_id)


@dataclass
class IndexFreeListKey(BaseIndexedDBKey):
  """An IndexedDB index free list key.

  Attributes:
    object_store_id: the ID of the object store containing the free list.
  """
  object_store_id: int

  def DecodeValue(self, decoder: utils.LevelDBDecoder):
    """Decodes the index free list value."""
    raise NotImplementedError('IndexFreeListKey.decode_value')

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
    ) -> IndexFreeListKey:
    """Decodes the index free list key."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.DatabaseMetaDataKeyType.INDEX_FREE_LIST:
      raise errors.ParserError('Not am IndexFreeListKey')
    _, object_store_id = decoder.DecodeVarint()
    return cls(
        offset=base_offset + offset, key_prefix=key_prefix,
        object_store_id=object_store_id)


@dataclass
class ObjectStoreNamesKey(BaseIndexedDBKey):
  """An IndexedDB object store name key.

  Attributes:
    object_store_name: the name of the object store.
  """
  object_store_name: str

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> int:
    """Decodes the object store names value."""
    return decoder.DecodeVarint()[1]

  @classmethod
  def FromDecoder(cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
                 base_offset: int = 0) -> ObjectStoreNamesKey:
    """Decodes the object store names key."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.DatabaseMetaDataKeyType.OBJECT_STORE_NAMES:
      raise errors.ParserError('Not am ObjectStoreNamesKey')
    _, object_store_name = decoder.DecodeStringWithLength()
    return cls(key_prefix=key_prefix, offset=base_offset + offset,
               object_store_name=object_store_name)


@dataclass
class IndexNamesKey(BaseIndexedDBKey):
  """An IndexedDB index name key.

  Attributes:
    index_name: the name of the index.
  """
  index_name: str

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> int:
    """Decodes the index names value."""
    return decoder.DecodeVarint()[1]

  @classmethod
  def FromDecoder(cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
                 base_offset: int = 0) -> IndexNamesKey:
    """Decodes the index names key."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.DatabaseMetaDataKeyType.INDEX_NAMES:
      raise errors.ParserError('Not am IndexNamesKey')
    _, index_name = decoder.DecodeStringWithLength()
    return cls(key_prefix=key_prefix, offset=base_offset + offset,
               index_name=index_name)


@dataclass
class DatabaseMetaDataKey(BaseIndexedDBKey):
  """An IndexedDB database metadata key.

  Attributes:
    metadata_type: the type of metadata that the key value contains.
  """
  metadata_type: definitions.DatabaseMetaDataKeyType

  def DecodeValue(
      self, decoder: utils.LevelDBDecoder) -> Union[str, int]:
    """Decodes the database metadata value."""
    if (self.metadata_type ==
        definitions.DatabaseMetaDataKeyType.ORIGIN_NAME):
      return decoder.DecodeString()[1]
    if (self.metadata_type ==
        definitions.DatabaseMetaDataKeyType.DATABASE_NAME):
      return decoder.DecodeString()[1]
    if (self.metadata_type ==
        definitions.DatabaseMetaDataKeyType.IDB_STRING_VERSION_DATA):
      return decoder.DecodeString()[1]
    if (self.metadata_type ==
        definitions.DatabaseMetaDataKeyType.MAX_ALLOCATED_OBJECT_STORE_ID):
      return decoder.DecodeVarint()[1]
    if (self.metadata_type ==
        definitions.DatabaseMetaDataKeyType.IDB_INTEGER_VERSION):
      return decoder.DecodeVarint()[1]
    if (self.metadata_type ==
        definitions.DatabaseMetaDataKeyType
            .BLOB_NUMBER_GENERATOR_CURRENT_NUMBER):
      return decoder.DecodeVarint()[1]
    raise errors.ParserError(
        f'Unknown database metadata type {self.metadata_type}')

  @classmethod
  def FromDecoder(
      cls,
      decoder: utils.LevelDBDecoder,
      key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> Union[
      DatabaseMetaDataKey,
      IndexFreeListKey,
      IndexMetaDataKey,
      IndexNamesKey,
      ObjectStoreFreeListKey,
      ObjectStoreMetaDataKey,
      ObjectStoreNamesKey]:
    """Decodes the database metadata key."""
    offset, metadata_value = decoder.PeekBytes(1)
    metadata_type = definitions.DatabaseMetaDataKeyType(metadata_value[0])

    if metadata_type in (
        definitions.DatabaseMetaDataKeyType.ORIGIN_NAME,
        definitions.DatabaseMetaDataKeyType.DATABASE_NAME,
        definitions.DatabaseMetaDataKeyType.IDB_STRING_VERSION_DATA,
        definitions.DatabaseMetaDataKeyType.MAX_ALLOCATED_OBJECT_STORE_ID,
        definitions.DatabaseMetaDataKeyType.IDB_INTEGER_VERSION,
        definitions.DatabaseMetaDataKeyType
            .BLOB_NUMBER_GENERATOR_CURRENT_NUMBER):
      return cls(key_prefix=key_prefix, offset=base_offset + offset,
                 metadata_type=metadata_type)
    if metadata_type == definitions.DatabaseMetaDataKeyType.INDEX_META_DATA:
      return IndexMetaDataKey.FromDecoder(
          decoder, key_prefix, base_offset)
    if (metadata_type ==
        definitions.DatabaseMetaDataKeyType.OBJECT_STORE_META_DATA):
      return ObjectStoreMetaDataKey.FromDecoder(
          decoder, key_prefix, base_offset)
    if (metadata_type ==
        definitions.DatabaseMetaDataKeyType.OBJECT_STORE_FREE_LIST):
      return ObjectStoreFreeListKey.FromDecoder(
          decoder, key_prefix, base_offset)
    if metadata_type == definitions.DatabaseMetaDataKeyType.INDEX_FREE_LIST:
      return IndexFreeListKey.FromDecoder(
          decoder, key_prefix, base_offset)
    if (metadata_type ==
        definitions.DatabaseMetaDataKeyType.OBJECT_STORE_NAMES):
      return ObjectStoreNamesKey.FromDecoder(
          decoder, key_prefix, base_offset)
    if metadata_type == definitions.DatabaseMetaDataKeyType.INDEX_NAMES:
      return IndexNamesKey.FromDecoder(
          decoder, key_prefix, base_offset)
    raise errors.ParserError(f'unknown database metadata type {metadata_type}.')


@dataclass
class ObjectStoreMetaDataKey(BaseIndexedDBKey):
  """An IndexedDB object store meta data key.

  Attributes:
    object_store_id: the ID of the object store.
    metadata_type: the object store metadata type.
  """
  object_store_id: int
  metadata_type: definitions.ObjectStoreMetaDataKeyType

  def DecodeValue(
      self, decoder: utils.LevelDBDecoder) -> Union[IDBKeyPath, str, bool, int]:
    """Decodes the object store metadata value."""
    if (self.metadata_type ==
        definitions.ObjectStoreMetaDataKeyType.OBJECT_STORE_NAME):
      return decoder.DecodeString()[1]
    if self.metadata_type == definitions.ObjectStoreMetaDataKeyType.KEY_PATH:
      return IDBKeyPath.FromDecoder(decoder)
    if (self.metadata_type ==
        definitions.ObjectStoreMetaDataKeyType.AUTO_INCREMENT_FLAG):
      return decoder.DecodeBool()[1]
    if (self.metadata_type ==
        definitions.ObjectStoreMetaDataKeyType.IS_EVICTABLE):
      return decoder.DecodeBool()[1]
    if (self.metadata_type ==
        definitions.ObjectStoreMetaDataKeyType.LAST_VERSION_NUMBER):
      return decoder.DecodeInt(signed=False)[1]
    if (self.metadata_type ==
        definitions.ObjectStoreMetaDataKeyType.MAXIMUM_ALLOCATED_INDEX_ID):
      return decoder.DecodeInt()[1]
    if (self.metadata_type ==
        definitions.ObjectStoreMetaDataKeyType.HAS_KEY_PATH):
      return decoder.DecodeBool()[1]
    if (self.metadata_type ==
        definitions.ObjectStoreMetaDataKeyType.KEY_GENERATOR_CURRENT_NUMBER):
      return decoder.DecodeInt()[1]
    raise errors.ParserError(f'Unknown metadata type {self.metadata_type}')

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> ObjectStoreMetaDataKey:
    """Decodes the object store metadata key."""
    offset, metadata_value = decoder.DecodeUint8()
    if (metadata_value !=
        definitions.DatabaseMetaDataKeyType.OBJECT_STORE_META_DATA):
      raise errors.ParserError('Not a ObjectStoreMetaDataKey')

    _, object_store_id = decoder.DecodeVarint()
    _, metadata_value = decoder.DecodeUint8()
    metadata_type = definitions.ObjectStoreMetaDataKeyType(metadata_value)
    return cls(
        offset=base_offset + offset, key_prefix=key_prefix,
        object_store_id=object_store_id, metadata_type=metadata_type)

@dataclass
class ObjectStoreDataValue:
  """The parsed values from an ObjectStoreDataKey.

  Attributes:
    unknown: an unknown integer (possibly a sequence number?).
    is_wrapped: True if the value was wrapped.
    blob_size: the blob size, only valid if wrapped.
    blob_offset: the blob offset, only valid if wrapped.
    value: the blink serialized value, only valid if not wrapped.
  """
  unkown: int
  is_wrapped: bool
  blob_size: Optional[int]
  blob_offset: Optional[int]
  value: Any


@dataclass
class ObjectStoreDataKey(BaseIndexedDBKey):
  """An IndexedDB object store data key.

  Attributes:
    encoded_user_key: the encoded user key.
  """
  encoded_user_key: IDBKey

  def DecodeValue(
      self, decoder: utils.LevelDBDecoder) -> ObjectStoreDataValue:
    """Decodes the object store data value."""
    _, unknown_integer = decoder.DecodeVarint()

    _, wrapped_header_bytes = decoder.PeekBytes(3)
    if len(wrapped_header_bytes) != 3:
      raise errors.DecoderError('Insufficient bytes')

    if (wrapped_header_bytes[0] == definitions.BlinkSerializationTag.VERSION and
        wrapped_header_bytes[1] == 0x11 and
        wrapped_header_bytes[2] == 0x01):
      _, blob_size = decoder.DecodeVarint()
      _, blob_offset = decoder.DecodeVarint()
      return ObjectStoreDataValue(
          unkown=unknown_integer,
          is_wrapped=True,
          blob_size=blob_size,
          blob_offset=blob_offset,
          value=None)
    _, blink_bytes = decoder.ReadBytes()
    blink_value = blink.V8ScriptValueDecoder.FromBytes(blink_bytes)
    return ObjectStoreDataValue(
        unkown=unknown_integer,
        is_wrapped=False,
        blob_size=None,
        blob_offset=None,
        value=blink_value)

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> ObjectStoreDataKey:
    """Decodes the object store data key."""
    if (key_prefix.GetKeyPrefixType() !=
        definitions.KeyPrefixType.OBJECT_STORE_DATA):
      raise errors.ParserError('Invalid KeyPrefix for ObjectStoreDataKey')
    offset = decoder.stream.tell()
    encoded_user_key = IDBKey.FromDecoder(decoder, offset)
    return cls(
        offset=base_offset + offset,
        key_prefix=key_prefix, encoded_user_key=encoded_user_key)


@dataclass
class ExistsEntryKey(BaseIndexedDBKey):
  """An IndexedDB exists entry key.

  Attributes:
    encoded_user_key: the encoded user key.
  """
  encoded_user_key: IDBKey

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> int:
    """Decodes the exists entry value."""
    _, data = decoder.ReadBytes()
    return int.from_bytes(data, byteorder='little', signed=False)

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> ExistsEntryKey:
    """Decodes the exists entry key."""
    offset = decoder.stream.tell()
    encoded_user_key = IDBKey.FromDecoder(decoder, offset)

    return cls(
        offset=base_offset + offset,
        key_prefix=key_prefix, encoded_user_key=encoded_user_key)


@dataclass
class IndexDataKey(BaseIndexedDBKey):
  """An IndexedDB data key.

  Attributes:
    encoded_user_key: the encoded user key.
    sequence_number: the sequence number of the data key.
    encoded_primary_key: the encoded primary key.
  """
  encoded_user_key: IDBKey
  sequence_number: Optional[int]
  encoded_primary_key: Optional[IDBKey]

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> Tuple[int, IDBKey]:
    """Decodes the index data key value."""
    _, version = decoder.DecodeVarint()
    idb_key = IDBKey.FromDecoder(decoder)
    return version, idb_key

  @classmethod
  def FromDecoder(cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
                 base_offset: int = 0) -> IndexDataKey:
    """Decodes the index data key."""
    offset = decoder.stream.tell()
    encoded_user_key = IDBKey.FromDecoder(decoder, offset)

    if decoder.NumRemainingBytes() > 0:
      _, sequence_number = decoder.DecodeVarint()
    else:
      sequence_number = None

    if decoder.NumRemainingBytes() > 0:
      encoded_primary_key_offset = decoder.stream.tell()
      encoded_primary_key = IDBKey.FromDecoder(
          decoder, encoded_primary_key_offset)
    else:
      encoded_primary_key = None

    return cls(
        key_prefix=key_prefix,
        encoded_user_key=encoded_user_key,
        sequence_number=sequence_number,
        encoded_primary_key=encoded_primary_key,
        offset=base_offset + offset)


@dataclass
class BlobEntryKey(BaseIndexedDBKey):
  """An IndexedDB blob entry key.

  Attributes:
    user_key: the user/primary key.
  """
  user_key: IDBKey

  def DecodeValue(
      self, decoder: utils.LevelDBDecoder) -> IndexedDBExternalObject:
    """Decodes the blob entry value."""
    return IndexedDBExternalObject.FromDecoder(decoder)

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> BlobEntryKey:
    """Decodes the blob entry key."""
    offset = decoder.stream.tell()
    user_key = IDBKey.FromDecoder(decoder, offset)

    return cls(key_prefix=key_prefix, user_key=user_key,
               offset=base_offset + offset)


@dataclass
class IndexedDbKey(BaseIndexedDBKey):
  """An IndexedDB key.

  A factory class for parsing IndexedDB keys.
  """

  METADATA_TYPE_TO_CLASS = {
      definitions.KeyPrefixType.GLOBAL_METADATA: GlobalMetaDataKey,
      definitions.KeyPrefixType.DATABASE_METADATA: DatabaseMetaDataKey,
      definitions.KeyPrefixType.OBJECT_STORE_DATA: ObjectStoreDataKey,
      definitions.KeyPrefixType.EXISTS_ENTRY: ExistsEntryKey,
      definitions.KeyPrefixType.INDEX_DATA: IndexDataKey,
      definitions.KeyPrefixType.BLOB_ENTRY: BlobEntryKey,
      definitions.KeyPrefixType.INVALID_TYPE: None
  }

  def DecodeValue(self, decoder: utils.LevelDBDecoder) -> Any:
    """Decodes the value from the current position of the LevelDBDecoder.

    Args:
      decoder: the stream decoder
    """

  @classmethod
  def FromDecoder(
      cls,
      decoder: utils.LevelDBDecoder,
      key_prefix: KeyPrefix,
      base_offset: int = 0
  ) -> Union[Type[DatabaseMetaDataKey], Type[ExistsEntryKey],
             Type[BlobEntryKey], Type[GlobalMetaDataKey],
             Type[IndexDataKey], Type[ObjectStoreDataKey]]:
    """Decodes the IndexedDB key."""
    key_type = key_prefix.GetKeyPrefixType()
    key_class = cls.METADATA_TYPE_TO_CLASS.get(key_type)
    if not key_class:
      raise errors.ParserError('Unknown KeyPrefixType')
    return key_class.FromDecoder(
        decoder=decoder,
        key_prefix=key_prefix,  #pytype: disable=bad-return-type
        base_offset=base_offset)


@dataclass
class IndexMetaDataKey(BaseIndexedDBKey):
  """An IndexedDB meta data key.

  Attributes:
    object_store_id: the ID of the object store.
    index_id: the index ID.
    metadata_type: the metadata key type.
  """
  object_store_id: int
  index_id: int
  metadata_type: definitions.IndexMetaDataKeyType

  def DecodeValue(
      self, decoder: utils.LevelDBDecoder) -> Union[bool, IDBKeyPath, str]:
    """Decodes the index metadata value."""
    if self.metadata_type == definitions.IndexMetaDataKeyType.INDEX_NAME:
      return decoder.DecodeString()[1]
    if self.metadata_type == definitions.IndexMetaDataKeyType.KEY_PATH:
      return IDBKeyPath.FromDecoder(decoder)
    if (self.metadata_type ==
        definitions.IndexMetaDataKeyType.MULTI_ENTRY_FLAG):
      return decoder.DecodeBool()[1]
    if self.metadata_type == definitions.IndexMetaDataKeyType.UNIQUE_FLAG:
      return decoder.DecodeBool()[1]
    raise errors.ParserError(
        f'Unknown index metadata type {self.metadata_type}')

  @classmethod
  def FromDecoder(
      cls,
      decoder: utils.LevelDBDecoder,
      key_prefix: KeyPrefix, base_offset: int = 0
  ) -> IndexMetaDataKey:
    """Decodes the index metadata key."""
    offset, key_type = decoder.DecodeUint8()
    if key_type != definitions.DatabaseMetaDataKeyType.INDEX_META_DATA:
      raise errors.ParserError('Not an IndexMetaDataKey.')
    _, object_store_id = decoder.DecodeVarint()
    _, index_id = decoder.DecodeVarint()
    _, metadata_bytes = decoder.ReadBytes(1)
    metadata_type = definitions.IndexMetaDataKeyType(metadata_bytes[0])
    return cls(offset=base_offset + offset, key_prefix=key_prefix,
               object_store_id=object_store_id, index_id=index_id,
               metadata_type=metadata_type)


@dataclass
class ExternalObjectEntry(utils.FromStreamMixin):
  """An IndexedDB external object entry.

  Args:
    offset: the offset of the ExternalObjectEntry.
    object_type: the external object type.
    blob_number: the blob number if the object type is a blob or file,
        None otherwise.
    mime_type: the mime type if the object type is a blob or file, None
        otherwise.
    size: the size if the object type is a blob or file, None otherwise.
    filename: the filename if the object is a blob or file, None otherwise.
    last_modified: the last modified time if a blob or file, None otherwise.
    token: the token if a filesystem access handle, None otherwise.
  """
  offset: int = field(compare=False)
  object_type: definitions.ExternalObjectType
  blob_number: Optional[int]
  mime_type: Optional[str]
  size: Optional[int]
  filename: Optional[str]
  last_modified: Optional[int]  # microseconds
  token: Optional[bytes]

  @classmethod
  def FromDecoder(
      cls,
      decoder: utils.LevelDBDecoder,
      base_offset: int = 0
  ) -> ExternalObjectEntry:
    """Decodes the external object entry."""
    offset, object_type_value = decoder.DecodeUint8()
    object_type = definitions.ExternalObjectType(object_type_value)
    if (object_type in
        (definitions.ExternalObjectType.BLOB,
         definitions.ExternalObjectType.FILE)):
      _, blob_number = decoder.DecodeVarint()
      _, mime_type = decoder.DecodeStringWithLength()
      _, size = decoder.DecodeVarint()
      if object_type == definitions.ExternalObjectType.FILE:
        _, filename = decoder.DecodeStringWithLength()
        _, last_modified = decoder.DecodeVarint()
      else:
        filename = None
        last_modified = None
      token = None
    elif (object_type ==
        definitions.ExternalObjectType.FILE_SYSTEM_ACCESS_HANDLE):
      blob_number = None
      mime_type = None
      size = None
      filename = None
      last_modified = None
      _, token = decoder.DecodeBlobWithLength()

    return cls(offset=base_offset + offset, object_type=object_type,
        blob_number=blob_number, mime_type=mime_type, size=size,
        filename=filename, last_modified=last_modified, token=token)


@dataclass
class IndexedDBExternalObject(utils.FromStreamMixin):
  """An IndexedDB external object.

  Args:
    offset: the offset of the IndexedDBExternalObject.
    entries: a list of external objects.
  """
  offset: int = field(compare=False)
  entries: list[ExternalObjectEntry]

  @classmethod
  def FromDecoder(
      cls,
      decoder: utils.LevelDBDecoder,
      base_offset: int = 0
  ) -> IndexedDBExternalObject:
    """Decodes the external object."""
    entries = []
    offset = decoder.stream.tell()
    while decoder.NumRemainingBytes():
      entry = ExternalObjectEntry.FromDecoder(decoder, base_offset)
      entries.append(entry)

    return cls(offset=base_offset + offset, entries=entries)


@dataclass
class ObjectStore:
  """An IndexedDB Object Store.

  Attributes:
    id: the object store ID.
    name: the object store name.
    records: the records in the object store.
  """
  id: int
  name: str
  records: list = field(default_factory=list, repr=False)


@dataclass
class IndexedDBRecord:
  """An IndexedDB Record.

  Attributes:
    offset: the offset of the record.
    key: the key of the record.
    value: the value of the record.
    sequence_number: if available, the sequence number of the record.
    type: the type of the record.
  """
  offset: int
  key: Any
  value: Any
  sequence_number: int
  type: int

  @classmethod
  def FromLevelDBRecord(
      cls, record: Union[ldb.LdbKeyValueRecord, log.ParsedInternalKey]
  ) -> IndexedDBRecord:
    """Returns an IndexedDBRecord from a ParsedInternalKey."""
    idb_key = IndexedDbKey.FromBytes(record.key, base_offset=record.offset)
    idb_value = idb_key.ParseValue(record.value)
    return cls(
      offset=record.offset,
      key=idb_key,
      value=idb_value,
      sequence_number=record.sequence_number if hasattr(
          record, 'sequence_number') else None,
      type=record.type)



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parsers for blink javascript serialized objects."""
import io
from typing import Any

from dfindexeddb import utils
from dfindexeddb.indexeddb import definitions
from dfindexeddb.indexeddb import v8



class V8ScriptValueDecoder:
  """A Blink V8 Serialized Script Value (SSV) decoder.

  Attributes:
    deserializer (v8.ValueDeserializer): the V8 value deserializer.
    raw_data (bytes): the raw bytes containing the serialized javascript
        value.
    version (int): the blink version.
  """
  _MIN_VERSION_FOR_SEPARATE_ENVELOPE = 16

  # As defined in trailer_reader.h
  _MIN_WIRE_FORMAT_VERSION = 21

  def __init__(self, raw_data: bytes):
    self.deserializer = None
    self.raw_data = raw_data
    self.version = 0

  def _ReadVersionEnvelope(self) -> int:
    """Reads the Blink version envelope.

    Returns:
      The number of bytes consumed reading the Blink version envelope or zero
          if no blink envelope is detected.
    """
    if not self.raw_data:
      return 0

    decoder = utils.StreamDecoder(io.BytesIO(self.raw_data))
    _, tag_value = decoder.DecodeUint8()
    tag = definitions.BlinkSerializationTag(tag_value)
    if tag != definitions.BlinkSerializationTag.VERSION:
      return 0

    _, version = decoder.DecodeUint32Varint()
    if version < self._MIN_VERSION_FOR_SEPARATE_ENVELOPE:
      return 0

    consumed_bytes = decoder.stream.tell()

    if version >= self._MIN_WIRE_FORMAT_VERSION:
      trailer_offset_data_size = 1 + 8 + 4 # 1 + sizeof(uint64) + sizeof(uint32)
      consumed_bytes += trailer_offset_data_size
      if consumed_bytes >= len(self.raw_data):
        return 0
    return consumed_bytes

  def ReadTag(self) -> definitions.BlinkSerializationTag:
    """Reads a blink serialization tag.

    Returns:
      The blink serialization tag.
    """
    _, tag_value = self.deserializer.decoder.DecodeUint8()
    return definitions.BlinkSerializationTag(tag_value)

  def ReadHostObject(self) -> Any:
    """Reads a host DOM object.

    Raises:
      NotImplementedError: when called.
    """
    tag = self.ReadTag()
    raise NotImplementedError(f'V8ScriptValueDecoder.ReadHostObject - {tag}')

  def Deserialize(self) -> Any:
    """Deserializes a Blink SSV.

    The serialization format has two 'envelopes'.
    [version tag] [Blink version] [version tag] [v8 version] ...

    Returns:
      The deserialized script value.
    """
    version_envelope_size = self._ReadVersionEnvelope()
    self.deserializer = v8.ValueDeserializer(
        io.BytesIO(self.raw_data[version_envelope_size:]), delegate=self)
    v8_version = self.deserializer.ReadHeader()
    if not self.version:
      self.version = v8_version
    return self.deserializer.ReadValue()

  @classmethod
  def FromBytes(cls, data: bytes) -> Any:
    """Deserializes a Blink SSV from the data array.

    Returns:
      The deserialized script value.
    """
    return cls(data).Deserialize()



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A CLI tool for dfindexeddb."""
import argparse
import dataclasses
from datetime import datetime
import json
import pathlib
import sys
import traceback

from dfindexeddb.leveldb import log
from dfindexeddb.leveldb import ldb
from dfindexeddb.indexeddb import chromium
from dfindexeddb import errors
from dfindexeddb.indexeddb import v8


class Encoder(json.JSONEncoder):
  """A JSON encoder class for dfindexeddb fields."""
  def default(self, o):
    if isinstance(o, bytes):
      return o.decode(encoding='ascii', errors='backslashreplace')
    if isinstance(o, datetime):
      return o.isoformat()
    if isinstance(o, v8.Undefined):
      return "<undefined>"
    if isinstance(o, v8.Null):
      return "<null>"
    if isinstance(o, set):
      return list(o)
    if isinstance(o, v8.RegExp):
      return str(o)
    return json.JSONEncoder.default(self, o)


def _Output(structure, to_json=False):
  """Helper method to output parsed structure to stdout."""
  if to_json:
    structure_dict = dataclasses.asdict(structure)
    print(json.dumps(structure_dict, indent=2, cls=Encoder))
  else:
    print(structure)


def IndexeddbCommand(args):
  """The CLI for processing a log/ldb file as indexeddb."""
  if args.source.name.endswith('.log'):
    records = list(
        log.LogFileReader(args.source).GetKeyValueRecords())
  elif args.source.name.endswith('.ldb'):
    records = list(
        ldb.LdbFileReader(args.source).GetKeyValueRecords())
  else:
    print('Unsupported file type.', file=sys.stderr)
    return

  for record in records:
    try:
      record = chromium.IndexedDBRecord.FromLevelDBRecord(record)
    except (errors.ParserError, errors.DecoderError) as err:
      print(
          (f'Error parsing blink value: {err} for {record.__class__.__name__} '
           f'at offset {record.offset}'), file=sys.stderr)
      print(f'Traceback: {traceback.format_exc()}', file=sys.stderr)
    _Output(record, to_json=args.json)


def LdbCommand(args):
  """The CLI for processing ldb files."""
  ldb_file = ldb.LdbFileReader(args.source)

  if args.structure_type == 'blocks':
    # Prints block information.
    for block in ldb_file.GetBlocks():
      _Output(block, to_json=args.json)

  elif args.structure_type == 'records':
    # Prints key value record information.
    for record in ldb_file.GetKeyValueRecords():
      _Output(record, to_json=args.json)


def LogCommand(args):
  """The CLI for processing log files."""
  log_file = log.LogFileReader(args.source)

  if args.structure_type == 'blocks':
    # Prints block information.
    for block in log_file.GetBlocks():
      _Output(block, to_json=args.json)

  elif args.structure_type == 'physical_records':
    # Prints log file physical record information.
    for log_file_record in log_file.GetPhysicalRecords():
      _Output(log_file_record, to_json=args.json)

  elif args.structure_type == 'write_batches':
    # Prints log file batch information.
    for batch in log_file.GetWriteBatches():
      _Output(batch, to_json=args.json)

  elif args.structure_type in ('parsed_internal_key', 'records'):
    # Prints key value record information.
    for record in log_file.GetKeyValueRecords():
      _Output(record, to_json=args.json)


def App():
  """The CLI app entrypoint."""
  parser = argparse.ArgumentParser(
      prog='dfindexeddb',
      description='A cli tool for the dfindexeddb package')

  parser.add_argument(
      '-s', '--source', required=True, type=pathlib.Path, 
      help='The source leveldb file')
  parser.add_argument('--json', action='store_true', help='Output as JSON')
  subparsers = parser.add_subparsers(required=True)

  parser_log = subparsers.add_parser('log')
  parser_log.add_argument(
      'structure_type', choices=[
          'blocks',
          'physical_records',
          'write_batches',
          'parsed_internal_key',
          'records'])
  parser_log.set_defaults(func=LogCommand)

  parser_log = subparsers.add_parser('ldb')
  parser_log.add_argument(
      'structure_type', choices=[
          'blocks',
          'records'])
  parser_log.set_defaults(func=LdbCommand)

  parser_log = subparsers.add_parser('indexeddb')
  parser_log.set_defaults(func=IndexeddbCommand)

  args = parser.parse_args()

  args.func(args)



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for dfindexeddb."""
import io
import os
import struct
from typing import BinaryIO, Tuple, Type, TypeVar


from dfindexeddb import errors


class StreamDecoder:
  """A helper class to decode primitive data types from BinaryIO streams.

   Attributes:
    stream (BinaryIO): the binary stream.
  """

  def __init__(self, stream: BinaryIO):
    """Initializes a StreamDecoder instance.

    Args:
      stream: the binary stream.
    """
    self.stream = stream

  def NumRemainingBytes(self) -> int:
    """Returns the number of bytes available to the decoder.

    Raises:
      errors.DecoderError: if there are a negative number of remaining bytes.
    """
    current_offset = self.stream.tell()
    end_offset = self.stream.seek(0, os.SEEK_END)
    self.stream.seek(current_offset, os.SEEK_SET)
    num_rem_bytes = end_offset - current_offset
    if num_rem_bytes < 0:
      raise errors.DecoderError('Negative number of remaining bytes.')
    return num_rem_bytes

  def ReadBytes(self, count: int = -1) -> Tuple[int, bytes]:
    """Reads a number of bytes or the remaining bytes from the binary stream.

    Args:
      count: the number of bytes to read, if this value is -1 or not provided,
          all the remaining bytes area read.

    Returns:
      the offset where bytes were read from and the bytes that were read.

    Raises:
      errors.DecoderError: if no bytes are available or there are not enough
          bytes to read.
    """
    offset = self.stream.tell()
    buffer = self.stream.read(count)
    if count == -1 and not buffer:
      raise errors.DecoderError(f'No bytes available at offset {offset}')
    if count != -1 and len(buffer) != count:
      raise errors.DecoderError(
          f'Read {len(buffer)} bytes, but wanted {count} at offset {offset}')
    return offset, buffer

  def PeekBytes(self, count: int) -> Tuple[int, bytes]:
    """Peeks a number of bytes from the binary stream.

    Args:
      count: the number of bytes to read, if this value is -1 or not provided,
          all the remaining bytes area read.

    Returns:
      the offset where bytes were read from and the bytes that were read.
    """
    offset, read_bytes = self.ReadBytes(count)
    self.stream.seek(offset, os.SEEK_SET)
    return offset, read_bytes

  def DecodeInt(
      self,
      byte_count: int = -1,
      byte_order: str = 'little',
      signed: bool = True
  ) -> Tuple[int, int]:
    """Decodes an integer from the binary stream.

    Args:
      byte_count: the number of bytes to read.
      byte_order: the endianness of the integer.
      signed: True if twos-complement is used to represent the integer.

    Returns:
      the decoded integer.
    """
    offset, buffer = self.ReadBytes(byte_count)
    return offset, int.from_bytes(buffer, byteorder=byte_order, signed=signed)

  def DecodeUint8(self) -> Tuple[int, int]:
    """Decodes an unsigned 8-bit integer from the binary stream."""
    return self.DecodeInt(byte_count=1, signed=False)

  def DecodeUint16(self) -> Tuple[int, int]:
    """Decodes an unsigned 16-bit integer from the binary stream."""
    return self.DecodeInt(byte_count=2, signed=False)

  def DecodeUint24(self) -> Tuple[int, int]:
    """Decodes an unsigned 24-bit integer from the binary stream."""
    return self.DecodeInt(byte_count=3, signed=False)

  def DecodeUint32(self) -> Tuple[int, int]:
    """Decodes an unsigned 32-bit integer from the binary stream."""
    return self.DecodeInt(byte_count=4, signed=False)

  def DecodeUint48(self) -> Tuple[int, int]:
    """Decodes an unsigned 48-bit integer from the binary stream."""
    return self.DecodeInt(byte_count=6, signed=False)

  def DecodeUint64(self) -> Tuple[int, int]:
    """Decodes an unsigned 64-bit integer from the binary stream."""
    return self.DecodeInt(byte_count=8, signed=False)

  def DecodeInt8(self) -> Tuple[int, int]:
    """Decodes a signed 8-bit integer from the binary stream."""
    return self.DecodeInt(byte_count=1, signed=True)

  def DecodeInt16(self) -> Tuple[int, int]:
    """Decodes a signed 16-bit integer from the binary stream."""
    return self.DecodeInt(byte_count=2, signed=True)

  def DecodeInt24(self) -> Tuple[int, int]:
    """Decodes a signed 24-bit integer from the binary stream."""
    return self.DecodeInt(byte_count=3, signed=True)

  def DecodeInt32(self) -> Tuple[int, int]:
    """Decodes a signed 32-bit integer from the binary stream."""
    return self.DecodeInt(byte_count=4, signed=True)

  def DecodeInt48(self) -> Tuple[int, int]:
    """Decodes a signed 48-bit integer from the binary stream."""
    return self.DecodeInt(byte_count=6, signed=True)

  def DecodeInt64(self) -> Tuple[int, int]:
    """Decodes a signed 64-bit integer from the binary stream."""
    return self.DecodeInt(byte_count=8, signed=True)

  def DecodeDouble(self, little_endian: bool = True) -> Tuple[int, float]:
    """Returns a Tuple of the offset and a double-precision float."""
    offset, blob = self.ReadBytes(8)
    if little_endian:
      value = struct.unpack('<d', blob)[0]
    else:
      value = struct.unpack('>d', blob)[0]
    return offset, value

  def DecodeFloat(self, little_endian: bool = True) -> Tuple[int, float]:
    """Returns a Tuple of the offset and a single-precision float."""
    offset, blob = self.ReadBytes(4)
    if little_endian:
      value = struct.unpack('<f', blob)[0]
    else:
      value = struct.unpack('>f', blob)[0]
    return offset, value

  def DecodeVarint(self, max_bytes: int = 10) -> Tuple[int, int]:
    """Returns a Tuple of the offset and the decoded base128 varint."""
    offset = self.stream.tell()
    varint = 0
    for i in range(0, max_bytes*7, 7):
      _, varint_part = self.ReadBytes(1)
      varint |= (varint_part[0] & 0x7f) << i
      if not varint_part[0] >> 7:
        break
    return offset, varint

  def DecodeZigzagVarint(self, max_bytes: int = 10) -> Tuple[int, int]:
    """Returns a Tuple of the offset and the decoded zigzag varint."""
    (offset, varint) = self.DecodeVarint(max_bytes)
    return offset, (varint >> 1) ^ -(varint & 1)

  def DecodeUint32Varint(self) -> Tuple[int, int]:
    """Decodes a variable unsigned 32-bit integer from the binary stream."""
    return self.DecodeVarint(max_bytes=5)

  def DecodeUint64Varint(self) -> Tuple[int, int]:
    """Decodes a variable unsigned 64-bit integer from the binary stream."""
    return self.DecodeVarint(max_bytes=10)

  def DecodeInt32Varint(self) -> Tuple[int, int]:
    """Decodes a variable signed 32-bit integer from the binary stream."""
    return self.DecodeZigzagVarint(max_bytes=5)

  def DecodeInt64Varint(self) -> Tuple[int, int]:
    """Decodes a variable signed 64-bit integer from the binary stream."""
    return self.DecodeZigzagVarint(max_bytes=10)



class LevelDBDecoder(StreamDecoder):
  """A helper class to decode data types from LevelDB files."""

  def DecodeBool(self) -> Tuple[int, bool]:
    """Returns a Tuple of the offset of decoding and the bool value."""
    offset, buffer = self.ReadBytes(1)
    return offset, buffer[0] is not None

  def DecodeString(self) -> Tuple[int, str]:
    """Returns a tuple of the offset of decoding and the string value.

    Raises:
      errors.DecoderError: when the parsed string buffer is not even (i.e.
          cannot be decoded as a UTF-16-BE string.
    """
    offset = self.stream.tell()
    buffer = self.stream.read()
    if len(buffer) % 2:
      raise errors.DecoderError(
          f'Odd number of bytes encountered at offset {offset}')
    return offset, buffer.decode('utf-16-be')

  def DecodeBlobWithLength(self) -> Tuple[int, bytes]:
    """Returns a tuple of a the offset of decoding and the binary blob."""
    offset, num_bytes = self.DecodeUint64Varint()
    _, blob = self.ReadBytes(num_bytes)
    return offset, blob

  def DecodeStringWithLength(self) -> Tuple[int, str]:
    """Returns a tuple of the offset of decoding and the string value."""
    offset, length = self.DecodeUint64Varint()
    _, buffer = self.ReadBytes(length*2)
    return offset, buffer.decode('utf-16-be')


T = TypeVar('T')


class FromStreamMixin:  # TODO: refactor leveldb parsers
  """A mixin for dataclasses parsing their attributes from a binary stream."""

  @classmethod
  def FromDecoder(
      cls: Type[T], decoder: LevelDBDecoder, base_offset: int = 0) -> T:
    """Decodes a class type from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The class instance.

    Raises:
      NotImplementedError if the child class does not implement this method.
    """
    raise NotImplementedError

  @classmethod
  def FromStream(
      cls: Type[T], stream: BinaryIO, base_offset: int = 0) -> T:
    """Decodes a class type from the current position of a binary stream.

    Args:
      stream: the binary stream.
      base_offset: the base offset of the binary stream.

    Returns:
      The class instance.
    """
    decoder = LevelDBDecoder(stream)
    return cls.FromDecoder(decoder, base_offset)

  @classmethod
  def FromBytes(
      cls: Type[T], raw_data: bytes, base_offset: int = 0) -> T:
    """Parses a class type from raw bytes.

    Args:
      raw_data: the raw data.
      base_offset: the base offset of the raw data.

    Returns:
      The class instance.
    """
    stream = io.BytesIO(raw_data)
    return cls.FromStream(stream=stream, base_offset=base_offset)
